#!BPY
"""
Name: 'XPPropertyEditor'
Blender: 249
Group: 'Object'
Tooltip: 'A less sanity-challenging properties editor'
"""
__author__ = "Alex Gifford"
__email__ = ""
__url__ = ""
__version__ = "1.00"
__bpydoc__ = """\

"""

import Blender
from Blender import *
from Blender import Draw, Image, Scene, BGL, Window, Material, Texture, Object, Mesh, Text, sys
from Blender.Draw  import *
from Blender.BGL import *
from Blender.Image import *
from XPlaneMacros import *
import math, struct, copy, string

# _______________________________________________________________________________________
state = 0

proceed = 0
g_proplists = []
g_index = 100
g_loc_x = 0
g_loc_y = 0
g_objgrouplists = []
g_ogindex		= 0
g_og_loc_x		= 0
g_og_loc_y		= 0
g_GUI_changed	= 0
g_undo_status	= 0			# 0: no data in registry, 1: undo data in registry, 2: "Undone" with redo data in registry
qualifs			= 0
g_sort_mode		= 0
g_search_flag   = 0
g_scope			= 0

g_mousexy		= [0, 0]
g_my_window_cds		= [0, 0, 0, 0]
windowtop		= 0
g_click_sel_prev	= [0, 0]
g_event_queue = [0,0]
g_confirmed = 0
myWindowID = Window.GetAreaID()

CL_BLACK		= [0.00, 0.00, 0.00]
CL_TEXT_DK		= [0.20, 0.20, 0.20]
CL_TEXT_MD		= [0.25, 0.25, 0.25]
CL_ACT_PINK		= [0.92, 0.85, 0.92]
CL_SEL_PINK		= [0.92, 0.65, 0.92]
CL_GREY75		= [0.75, 0.75, 0.75]
CL_GREY70		= [0.70, 0.70, 0.70]
CL_MACRO		= [0.83, 0.83, 0.83]
CL_RED			= [1.00, 0.34, 0.22]
CL_ORANGE		= [1.00, 0.76, 0.24]
CL_CYAN			= [0.45, 0.89, 0.75]

def getMyWindowID():
	global	myWindowID
	return myWindowID

# ___________________________________________________________________________________________________________________________________________________________________________________________

class PropObject:

	def __init__(self):
		self.mylist = None
		self.index = 0
		self.object = None
		self.prop = None
		self.types = {'BOOL':0, 'INT':1, 'FLOAT':2, 'STRING':3}
		self.type_enums = {0:'BOOL', 1:'INT', 2:'FLOAT', 3:'STRING'}
		self.type = 0
		self.name = ""
		self.value_bool = 1
		self.value_int = 0
		self.value_float = 0.0
		self.value_string = ""
		self.value = {'BOOL':self.value_bool, 'INT':self.value_int, 'FLOAT':self.value_float, 'STRING':self.value_string}
		self.value_macro = ""
		self.macro_index = 0
		self.macro_menu_string = ""
		self.event = 0
		self.nameEvent = 0
		self.typeEvent = 0
		self.valueEvent = 0
		self.delEvent = 0
		self.isFocusEvent = 0
		self.showMacroEvent = 0
		self.focusNameEvent = 0
		self.focusValueEvent = 0
		self.obj_bt_move_up = None
		self.obj_mn_macro_index = 0
		self.loc_x = 0
		self.loc_y = 0
		self.offset_y = 0
		self.prop_has_macro = 0
		self.show_macro = 0
		self.is_focus = 0

	def setUp(self, parent, index, property, event, obj):
		self.mylist = parent
		self.object = obj
		self.prop = property
		self.index = index
		self.name = property.name
		self.type = self.types[property.type]
		self.value[property.type] = property.data
		if self.type == 0:
			self.value_bool = self.prop.data

		elif self.type == 1:
			self.value_int = self.prop.data

		elif self.type == 2:
			self.value_float = self.prop.data

		elif self.type == 3:
			self.value_string = self.prop.data

		self.event				= event
		self.typeEvent			= event + 1
		self.nameEvent			= event + 2
		self.valueBoolEvent		= event + 3
		self.value2Event		= event + 4
		self.valueIntEvent		= event + 5
		self.valueFloatEvent	= event + 6
		self.valueStringEvent	= event + 7
		self.delEvent			= event + 8
		self.isFocusEvent		= event + 9
		self.showMacroEvent		= event + 10
		self.macroKeyEvent		= event + 11
		self.focusNameEvent		= event + 12
		self.focusValueEvent	= event + 13

		self.obj_bt_delete		= Create(0)
		self.obj_tg_is_focus	= Create(0)
		self.obj_bt_focus_name	= Create(0)
		self.obj_bt_focus_value	= Create(0)
		self.obj_tg_show_macro	= Create(0)
		self.obj_mn_type		= Create(self.types[self.prop.type])
		self.obj_str_name		= Create(self.prop.name)
		self.obj_value_bool		= Create(self.value_bool)
		self.obj_value_int		= Create(self.value_int)
		self.obj_value_float	= Create(self.value_float)
		self.obj_value_string	= Create(self.value_string)
		self.obj_value_notbool	= Create(1 - self.value_bool)
		self.obj_mn_macro_index	= Create(self.macro_index)

		if not macros:
			PupMenu("CAUTION: Can't find file 'XPlaneMacros.py'")
			self.macro_menu_string = ""
		else:
			for i, m in enumerate(macros):
				self.macro_menu_string += ("|%s %%x%d" % (m[0], i))
		if 'MACRO' in self.name.upper() and macros:
			self.prop_has_macro = 1
			if self.value_string:
				for i, [k, v] in enumerate(macros):
					if k == self.value_string:
						self.macro_index = i
						self.obj_mn_macro_index.val = i
						self.value_macro = v
		else:
			self.prop_has_macro = 0


	def setValue(self, inVal):

		if isinstance(inVal, int):
			if self.type == 0:
				self.value_bool = inVal
				self.value['BOOL'] = inVal
			elif self.type == 1:
				self.value_int = inVal
				self.value['INT'] = inVal

		elif isinstance(inVal, float):
			self.value_float = inVal
			self.type = 2
			self.value['FLOAT'] = inVal

		elif isinstance(inVal, str):
			self.value_string = inVal
			self.type = 3
			self.value['STRING'] = inVal

	def getValue(self):

		if self.type == 0:
			return self.value_bool

		elif self.type == 1:
			return self.value_int

		elif self.type == 2:
			return self.value_float

		elif self.type == 3:
			return self.value_string

	def setType(self, inType):
		if isinstance(inType, str):		# for when a blender 'type' string is passed in
			self.type = self.types[inType]
		elif isinstance(inType, int):	# for when an int 'type' enum is passed in
			self.type = inType

		# initialize all data values. This is inline with Blender behaviour:
		self.value_bool = 0
		self.value_int = 0
		self.value_float = 0.0
		self.value_string = ""
		self.value['BOOL'] = 0
		self.value['INT'] = 0
		self.value['FLOAT'] = 0.0
		self.value['STRING'] = ""

	def getType(self):
		return self.type_enums[self.type]

	def refreshBlenderProps(self):
		props = self.object.getAllProperties()
		details = []
		for p in props:
			if p == self.prop:
				details.append([self.name, self.value[self.type_enums[self.type]], self.type_enums[self.type]])
			else:
				details.append([p.name, p.data, p.type])
		self.object.removeAllProperties()
		for d in details:
			self.object.addProperty(d[0], d[1], d[2])

		self.mylist.writeAllPropsToBlender()

	def handleEvent(self, event):

		global g_prop_focus

		if event == self.typeEvent:
			self.setType(self.obj_mn_type.val)

			self.refreshBlenderProps()
			Draw()

		elif event == self.nameEvent:

			self.prop.setName(self.obj_str_name.val)
			self.name = self.prop.name
			if 'MACRO' in self.name.upper() and macros:
				self.prop_has_macro = 1
				if self.value_string:
					for i, [k, v] in enumerate(macros):
						if k == self.value_string:
							self.macro_index = i
							self.obj_mn_macro_index.val = i
							self.value_macro = macros[i][1]
			Window.RedrawAll()

		elif event == self.valueBoolEvent:
			self.prop.data = self.obj_value_bool.val
			self.value_bool = self.obj_value_bool.val
			self.value['BOOL'] = self.obj_value_bool.val
			self.refreshBlenderProps()
			Draw()
		elif event == self.value2Event:
			self.prop.data = 1 - self.obj_value_notbool.val
			self.value_bool = 1 - self.obj_value_notbool.val
			self.value['BOOL'] = 1 - self.obj_value_bool.val
			self.refreshBlenderProps()
			Draw()
		elif event == self.valueIntEvent:
			self.prop.data = self.obj_value_int.val
			self.value['INT'] = self.obj_value_int.val
			self.value_int = self.value['INT']
			Window.RedrawAll()
		elif event == self.valueFloatEvent:
			self.prop.data = self.obj_value_float.val
			self.value['FLOAT'] = self.obj_value_float.val
			self.value_float = self.value['FLOAT']
			Window.RedrawAll()
		elif event == self.valueStringEvent:
			self.prop.data = self.obj_value_string.val
			self.value['STRING'] = self.obj_value_string.val
			self.value_string = self.value['STRING']
			Window.RedrawAll()
		elif event == self.delEvent:

			if qualifs & Window.Qual.SHIFT:
				if queryAction(self.delEvent):
					removePropFromSelected(self.name)

			elif qualifs & Window.Qual.CTRL:
				if queryAction(self.delEvent):
					copyPropToSelected(self.prop.name, self.prop.data, self.prop.type)

			else:
				if queryAction(self.delEvent):
					self.object.removeProperty(self.name)
					showResults("The property was removed.")
					buildTree(1)
			Draw()
		elif event == self.macroKeyEvent:
			self.macro_index = self.obj_mn_macro_index.val
			self.value_string = macros[self.macro_index][0]
			self.value['STRING'] = self.value_string
			self.prop.data = self.value_string
			self.value_macro = macros[self.macro_index][1]
			Draw()
		elif event == self.showMacroEvent:
			self.show_macro = self.obj_tg_show_macro.val
			Draw()
		elif event == self.isFocusEvent:
			self.is_focus = self.obj_tg_is_focus.val
			resetFocus(self.index)

		elif event == self.focusNameEvent:
			g_prop_focus.setUp(self.mylist, self.index, self.prop,	self.event, self.object)
			buildFocusTree(1)

		elif event == self.focusValueEvent:
			g_prop_focus.setUp(self.mylist, self.index, self.prop,	self.event, self.object)
			buildFocusTree(2)

	def draw(self, offsetY, maxY):

		global g_GUI_changed
		self.offset_y = offsetY
		dy = self.offset_y + self.loc_y
		if dy > maxY:
			return self.offset_y

		if self.show_macro and self.value_macro:
			lines = self.value_macro.splitlines()
			BGL.glColor3f(*CL_MACRO)
			BGL.glRecti(self.loc_x + 1  , dy + 21, self.loc_x + 804, dy - 22 * len(lines))

			BGL.glColor3f(*CL_TEXT_MD)
			for i, l in enumerate(lines):
				BGL.glRasterPos2i( self.loc_x + 12, self.loc_y + 7 +self.offset_y)
				self.offset_y -= 22
				Text(l, 'small')

		dy = self.offset_y + self.loc_y
		if qualifs & Window.Qual.CTRL:
			butname = "Copy"
			tipbut  = "Copy this property to all selected objects in visible layers of this scene."
		elif qualifs & Window.Qual.SHIFT:
			butname = "Del Sel"
			tipbut  = "Delete this property from all selected objects in visible layers of this scene."
		else:
			butname = "Delete"
			tipbut  = "Delete this property from this object."
		self.obj_bt_delete				= PushButton(	butname,			self.delEvent,			self.loc_x,		  dy,	60, 22,														tipbut)
		self.obj_bt_focus_name			= PushButton(	"fc",				self.focusNameEvent,	self.loc_x + 320, dy,	22, 22,		"Focus on all instances of like-named properties within the selected scope.")
		self.obj_bt_focus_value			= PushButton(	"fc",				self.focusValueEvent,	self.loc_x + 782, dy,	22, 22,		"Focus on all instances of like-named properties with this value within the selected scope.")

		menutypename = "Types %t|Bool %x0|Int %x1|Float %x2|String %x3"
		self.obj_mn_type				= Menu(			menutypename,		self.typeEvent,			self.loc_x + 60,  dy,	60,  22,	self.type,										"")
		self.obj_str_name				= String(		"Name: ",			self.nameEvent,			self.loc_x + 120, dy,	200, 22,	self.name,				31,						"")
		if self.type == 0:
			self.obj_value_bool			= Toggle(		"True",				self.valueBoolEvent,	self.loc_x + 342, dy,	220, 22,	self.value_bool,								"")
			self.obj_value_notbool		= Toggle(		"False",			self.value2Event,		self.loc_x + 562, dy,	220, 22,	not self.value_bool,							"")
		if self.type == 1:
			self.obj_value_int			= Number(		"",					self.valueIntEvent,		self.loc_x + 342, dy,	440, 22,	self.value_int,			-100000,	100000,		"")
		elif self.type == 2:
			self.obj_value_float		= Number(		"",					self.valueFloatEvent,	self.loc_x + 342, dy,	440, 22,	self.value_float,		-100000.0,	100000.0,	"")
		elif self.type == 3:
			if self.prop_has_macro:
				self.obj_tg_show_macro	= Toggle(		"Expand Macro",		self.showMacroEvent,	self.loc_x + 342, dy,	100, 22,	self.show_macro,								"Show contents of selected X-Plane macro.")
				self.obj_mn_macro_index	= Menu(self.macro_menu_string,		self.macroKeyEvent,		self.loc_x + 442, dy,	340, 22,	self.macro_index,								"")
			else:
				self.obj_value_string	= String(		"",					self.valueStringEvent,	self.loc_x + 342, dy,	440, 22,	self.value_string,		399,					"")

		return self.offset_y

# ___________________________________________________________________________________________________________________________________________________________________________________________

class ObPropList:

	def __init__(self, index, object, depth, x, y):

		self.obj				= object
		self.name				= object.name
		self.obj_st_obj_name	= Create(self.name)
		obj_bt_add_prop			= Create(0)
		obj_bt_make_active		= Create(0)
		self.ID					= index
		self.depth				= depth
		self.event_add			= self.ID * 10
		self.event_name			= self.event_add + 1
		self.event_make_active  = self.event_add + 2
		self.loc_x				= x
		self.loc_y				= y - 22
		props					= object.getAllProperties()
		self.prop_objects		= []

		for i, prop in enumerate(props):
			p = PropObject()
			p.setUp(self, index+1+i, prop,  (index+1+i) * 20, object)
#			p.setData(index+1+i, prop.name, prop.type, prop.data,  (index+1+i) * 20, object)
			p.loc_x = self.loc_x
			p.loc_y = self.loc_y - 22 * (i+1)
			self.prop_objects.append(p)
		self.prop_count			= len(self.prop_objects)

	def handleEvent(self, event):

		for po in self.prop_objects:
			po.handleEvent(event)
		global g_proplists
		if event == self.event_add:
			self.addProperty()

		if event == self.event_name:
			self.name = self.obj_st_obj_name.val
			self.obj.setName(self.name)

		if event == self.event_make_active:

			if self.obj.sel:
				if self.obj == g_ob_active:
					if qualifs & Window.Qual.SHIFT:
						self.obj.select(False)
					else:
						Blender.Scene.GetCurrent().objects.selected = [self.obj]
	#					for pl in [pl for pl in g_proplists if pl is not self]:
	#						pl.obj.select(False)
	#					self.obj.select(True)
				else:
					if qualifs & Window.Qual.SHIFT:
						setActiveObj(self.obj)
					else:
						self.obj.select(False)

			elif not self.obj.sel:
				self.obj.select(True)
				setActiveObj(self.obj)
				if not (qualifs & Window.Qual.SHIFT):
					Blender.Scene.GetCurrent().objects.selected = [self.obj]
#					for pl in [pl for pl in g_proplists if pl is not self]:
#						pl.obj.select(False)

			Window.RedrawAll()

	def writeAllPropsToBlender(self):

		self.obj.removeAllProperties()
		for po in self.prop_objects:
			self.obj.addProperty(po.name, po.value[po.type_enums[po.type]], po.type_enums[po.type])
			po.prop = self.obj.getProperty(po.name)
		buildTree(refresh_props_only=1)

	def selectChildren(self):
		if self.obj.sel:
			kids = getChildren(self.obj, Blender.Scene.GetCurrent().objects)
			for c in kids:
				c.select(True)
			buildTree()
		gui()
		Draw()

	def drawList(self, offsetY, maxY):

		try:
			list_offset_y = offsetY

			max_y = maxY - 22 * self.depth
			list_offset_y = offsetY
			dy = list_offset_y + self.loc_y
			header_offset_y = min(dy, max_y)
			list_offset_y += ((header_offset_y - dy)%22)
			dy = list_offset_y + self.loc_y

			BGL.glColor3f(*CL_GREY75)
			BGL.glRecti(		self.loc_x + 1,		header_offset_y + 1, self.loc_x + 442, header_offset_y + 22)


			if self.obj:
				if self.obj.isSelected():
					if self.obj == g_ob_active:
						BGL.glColor3f(*CL_ACT_PINK)
						BGL.glRecti(self.loc_x + 102,	header_offset_y + 1, self.loc_x + 120, header_offset_y + 22)
					else:
						BGL.glColor3f(*CL_SEL_PINK)
						BGL.glRecti(self.loc_x + 102,	header_offset_y + 1, self.loc_x + 120, header_offset_y + 22)
				else:
					BGL.glColor3f(*CL_GREY70)
					BGL.glRecti(	self.loc_x + 102 ,	header_offset_y + 1, self.loc_x + 120, header_offset_y + 22)

			BGL.glColor3f(*CL_TEXT_MD)
			BGL.glRasterPos2i(	self.loc_x + 348,	header_offset_y + 6)
			Text("Type: " )

			BGL.glRasterPos2i(	self.loc_x + 382,	header_offset_y + 6)
			Text(self.obj.type, 'normal')

			BGL.glRasterPos2i(	self.loc_x + 63,	header_offset_y + 6)
			Text("Props   %d" % self.prop_count, 'normal')

			BGL.glRasterPos2i(	self.loc_x + 6,		header_offset_y + 6)
			Text("Depth %d"  % self.depth)

	#		quals = Window.GetKeyQualifiers()

			butselname = "Select"
			tipsel = "Select object."
			if self.obj.sel:
				if qualifs & Window.Qual.SHIFT:
					if self.obj == g_ob_active:
						butselname	= "Deselect"
						tipsel		= "Deselect object."
					else:
						butselname	= "Make Active"
						tipsel		= "Make this the active object."
				elif self.obj != g_ob_active:
					butselname		= "Deselect"
					tipsel			= "Deselect object."

			elif qualifs & Window.Qual.SHIFT and Blender.Scene.GetCurrent().objects.selected:
				butselname			= " + Select   "
				tipsel				= "Add this object to the current selection."

			self.obj_st_obj_name =		String(		"OB: ",				self.event_name,		self.loc_x + 120 ,	header_offset_y,	222, 22,	self.name,	21,	"")
			self.obj_bt_add_prop =		PushButton( "Add Property",		self.event_add,			self.loc_x + 542,	header_offset_y,	262, 22,					"")
			self.obj_bt_make_active =	PushButton( butselname,			self.event_make_active,	self.loc_x + 442,	header_offset_y,	100, 22,					tipsel)

			for po in self.prop_objects:
				list_offset_y = po.draw(list_offset_y,  header_offset_y - 22)

			return list_offset_y

		except RuntimeError:
			return list_offset_y

	def addProperty(self):
		self.obj.addProperty("name", "", "STRING")
		buildTree(1)

# ___________________________________________________________________________________________________________________________________________________________________________________________

class ObjGroupObject:

	def __init__(self):
		self.data = []
		self.my_list = None
		self.index = 0
		self.objs = []
		self.obj_count = 0
		self.types = {'BOOL':0, 'INT':1, 'FLOAT':2, 'STRING':3}
		self.type_enums = {0:'BOOL', 1:'INT', 2:'FLOAT', 3:'STRING'}
		self.type = 3
		self.typestring = 'STRING'
		self.name = ""
		self.value_bool = 1
		self.value_int = 0
		self.value_float = 0.0
		self.value_string = ""
		self.value_macro = ""
		self.value = {'BOOL':self.value_bool, 'INT':self.value_int, 'FLOAT':self.value_float, 'STRING':self.value_string}
		self.value_any = 0
		self.macro_index = 0
		self.macro_menu_string = ""
		self.event = 0
		self.nameEvent = 0
		self.valueEvent = 0
		self.deleteEvent = 0
		self.showMacroEvent = 0
		self.moveEvent = 0
		self.selectEvent = 0
		self.obj_bt_select = None
		self.obj_bt_move = None
		self.obj_bt_delete = None
		self.obj_mn_macro_index = 0
		self.loc_x = 0
		self.loc_y = 0
		self.offset_y = 0
		self.prop_has_macro = 0
		self.show_macro = 0
		self.is_focus = 0
		self.selected = 0
		self.active = 0

	def setData2(self, index, prop, event, obj):
		self.index = index
		self.name = prop.name
		self.type = self.types[prop.type]
		self.typestring = prop.type
		self.objs = []
		self.objs.append(obj)
		self.obj_count = 1
		if self.type == 0:
			self.value_bool = prop.data
			self.value['BOOL'] = prop.data
		elif self.type == 1:
			self.value_int = prop.data
			self.value['INT'] = prop.data
		elif self.type == 2:
			self.value['FLOAT'] = prop.data
			self.value_float = prop.data
		elif self.type == 3:
			self.value_string = prop.data
			self.value['STRING'] = prop.data

		self.value_any = prop.data

		self.obj_bt_delete		= Create(0)
		self.obj_bt_move		= Create(0)
		self.obj_bt_select		= Create(0)
		self.obj_tg_show_macro	= Create(0)
		self.obj_mn_type		= Create(self.type)
		self.obj_str_name		= Create(self.name)
		self.obj_value_bool		= Create(self.value_bool)
		self.obj_value_int		= Create(self.value_int)
		self.obj_value_float	= Create(self.value_float)
		self.obj_value_string	= Create(self.value_string)
		self.obj_value_notbool	= Create(1 - self.value_bool)
		self.obj_mn_macro_index	= Create(self.macro_index)
		self.event				= event
		self.typeEvent			= event + 1
		self.nameEvent			= event + 2
		self.valueEvent			= event + 3
		self.value2Event		= event + 4
		self.deleteEvent		= event + 5
		self.selectEvent		= event + 6
		self.showMacroEvent		= event + 7
		self.macroKeyEvent		= event + 8
		self.moveEvent			= event + 9

		if not macros:
			PupMenu("WARNING: Can't find file 'XPlaneMacros.py'")
			self.macro_menu_string = ""
		else:
			for i, m in enumerate(macros):
				self.macro_menu_string += ("|%s %%x%d" % (m[0], i))
		if 'MACRO' in self.name.upper() and macros:
			self.prop_has_macro = 1
			# Check to see if the given macro string (i.e. the actual macro KEY string) exists in the list of macros. If yes, find its index.
			if self.value_string:
				for i, [k, v] in enumerate(macros):
					if k == self.value_string:
						self.macro_index = i
						self.obj_mn_macro_index.val = i
						# Get the value string of the macro:
						self.value_macro = v
		else:
			self.prop_has_macro = 0

	def setData(self, index, name, type, value, event, objects):

		self.index = index
		self.name = name
		self.type = self.types[type]
		self.typestring = type
		self.objs = objects
		self.obj_count = len(self.objs)
		if self.type == 0:
			self.value_bool =value
			self.value['BOOL'] =value
		elif self.type == 1:
			self.value_int = value
			self.value['INT'] = value
		elif self.type == 2:
			self.value['FLOAT'] =value
			self.value_float = value
		elif self.type == 3:
			self.value_string =value
			self.value['STRING'] =value
		self.value_any = value

		self.obj_bt_delete		= Create(0)
		self.obj_bt_move		= Create(0)
		self.obj_bt_select		= Create(0)
		self.obj_tg_show_macro	= Create(0)
		self.obj_mn_type		= Create(self.type)
		self.obj_str_name		= Create(self.name)
		self.obj_value_bool		= Create(self.value_bool)
		self.obj_value_int		= Create(self.value_int)
		self.obj_value_float	= Create(self.value_float)
		self.obj_value_string	= Create(self.value_string)
		self.obj_value_notbool	= Create(1 - self.value_bool)
		self.obj_mn_macro_index	= Create(self.macro_index)
		self.event				= event
		self.typeEvent			= event + 1
		self.nameEvent			= event + 2
		self.valueEvent			= event + 3
		self.value2Event		= event + 4
		self.deleteEvent		= event + 5
		self.selectEvent		= event + 6
		self.showMacroEvent		= event + 7
		self.macroKeyEvent		= event + 8
		self.moveEvent			= event + 9


		if not macros:
			PupMenu("WARNING: Can't find file 'XPlaneMacros.py'")
			self.macro_menu_string = ""
		else:
			for i, m in enumerate(macros):
				self.macro_menu_string += ("|%s %%x%d" % (m[0], i))
		if 'MACRO' in name.upper() and macros:
			self.prop_has_macro = 1
			# Check to see if the given macro string (i.e. the actual macro KEY string) exists in the list of macros. If yes, find its index.
			if self.value_string:
				for i, [k, v] in enumerate(macros):
					if k == self.value_string:
						self.macro_index = i
						self.obj_mn_macro_index.val = i
						# Get the value string of the macro:
						self.value_macro = v
		else:
			self.prop_has_macro = 0

	def addObject(self, object):
		self.objs.append(object)
		self.obj_count += 1

	def getObjectCount(self):
		self.obj_count = len(self.objs)
		return self.obj_count

	def setValue(self, inVal):
		if isinstance(inVal, int):
			if self.type == 0:
				self.value_bool = inVal
				self.value['BOOL'] = inVal
			elif self.type == 1:
				self.value_int = inVal
				self.value['INT'] = inVal

		elif isinstance(inVal, float):
			self.value_float = inVal
			self.type = 2
			self.value['FLOAT'] = inVal

		elif isinstance(inVal, str):
			self.value_string = inVal
			self.type = 3
			self.value['STRING'] = inVal
		self.value_any = inVal

	def getValue(self):

		if self.type == 0:
			return self.value_bool

		elif self.type == 1:
			return self.value_int

		elif self.type == 2:
			return self.value_float

		elif self.type == 3:
			return self.value_string

	def getValueAny(self):
		return self.value_any

	def updateData(self, searchScope, replaceMode, searchTerm, replaceTerm):
		""" This updates the objectgroupobject name, type and value,
			for example, to reflect the results of a search and replace
		"""
		oldname = self.name
		changed = 0
		newdata = None

		if self.type == 0:
			olddata = self.value_bool
		elif self.type == 1:
			olddata = self.value_int
		elif self.type == 2:
			olddata = self.value_float
		elif self.type == 3:
			olddata = self.value_string


		if searchScope == 0:			# names
			if replaceMode == 0:		# substring
				newname = 	smartReplace(self.name, str(searchTerm), str(replaceTerm), G_CASE_SENSITIVE)
			elif replaceMode == 1:
				newname = str(replaceTerm)
			if newname != self.name and len(newname) <= 31:
				self.name = newname

		elif searchScope == 1:			# values

			if self.type == 0:
				if replaceTerm in ['TRUE', 'YES', '1', 1, 1.0, '']:
					self.value_bool = 1
				else:
					self.value_bool = 0

			elif self.type == 1:
				if isIntString(replaceTerm):
					if not replaceMode:
						newdata = int(str(self.value_int).replace(str(searchTerm), str(replaceTerm)))
					else:
						newdata = int(replaceTerm)
				else:
					newdata = self.value_int
				self.value_int = newdata

			elif self.type == 2:
				if isFloatString(replaceTerm):
					if not replaceMode:
						newdata = float(str(self.value_float).replace(str(searchTerm), str(replaceTerm)))
					else:
						newdata = float(replaceTerm)
				else:
					newdata = self.value_float
				self.value_float = newdata

			elif self.type == 3:
				if not replaceMode:
					newdata = smartReplace(self.value_string, str(searchTerm), str(replaceTerm), G_CASE_SENSITIVE)
				else:
					newdata = replaceTerm
				self.value_string = newdata

			self.value_any = newdata

		for ob in self.objs:
			props = ob.getAllProperties()
			details = []
			for p in props:
				n = p.name
				d = p.data
				t = p.type
				if searchScope == 0:
					if n == oldname:
						n = self.name
						changed += 1
				elif searchScope == 1:
					if d == olddata:
						d = newdata
						t = self.typestring
						changed += 1
				details.append([n, d, t])
			ob.removeAllProperties()
			for d in details:
				ob.addProperty(d[0], d[1], d[2])

		return changed

	def refreshBlenderProps(self):

		for ob in self.objs:
			details = []
			for p in ob.getAllProperties():
				if p.name == self.name:
					details.append([p.name, self.value[self.type_enums[self.type]], self.type_enums[self.type]])
				else:
					details.append([p.name, p.data, p.type])
			ob.removeAllProperties()
			for d in details:
				ob.addProperty(d[0], d[1], d[2])
			Window.RedrawAll()

	def matchesProp(self, inProp):

		if inProp.name != self.name or self.types[inProp.type] != self.type:
			return False
		else:
			if self.type == 0:
				dat = self.value_bool
			elif self.type == 1:
				dat = self.value_int
			elif self.type == 2:
				dat = self.value_float
			elif self.type == 3:
				dat = self.value_string
			else:
				dat = None

			if dat == inProp.data:
				return True
			else:
				return False

	def handleEvent(self, event):

		global g_mode, g_prop_focus, g_ob_active, g_click_sel_prev, g_mousexy

		if event == self.nameEvent:
			oldname = self.name
			self.name = self.obj_str_name.val
			if 'MACRO' in self.name.upper() and macros:
				self.prop_has_macro = 1
				if self.value_string:
					for i, [k, v] in enumerate(macros):
						if k == self.value_string:
							self.macro_index = i
							self.obj_mn_macro_index.val = i
							self.value_macro = macros[i][1]

			for ob in self.objs:
				for p in ob.getAllProperties():
					if p.name == oldname:
						p.name = self.name
			self.refreshBlenderProps()

		elif event == self.valueEvent:

			if self.type == 0:
				self.setValue(self.obj_value_bool.val)
			elif self.type == 1:
				self.setValue(self.obj_value_int.val)
			elif self.type == 2:
				self.setValue(self.obj_value_float.val)
			elif self.type == 3:
				self.setValue(self.obj_value_string.val)
			self.refreshBlenderProps()

		elif event == self.value2Event:
			self.setValue(1 - self.obj_value_notbool.val)
			self.refreshBlenderProps()

		elif event == self.deleteEvent:

			if qualifs & Window.Qual.CTRL:

				if queryAction(self.deleteEvent, "Do you really want to copy this property to all selected objects?" ):
					if self.type == 0:
						copyPropToSelected(self.name, self.value_bool, 'BOOL')
					elif self.type == 1:
						copyPropToSelected(self.name, self.value_int, 'INT')
					elif self.type == 2:
						copyPropToSelected(self.name, self.value_float, 'FLOAT')
					elif self.type == 3:
						copyPropToSelected(self.name, self.value_string, 'STRING')
			else:
				if queryAction(self.deleteEvent, "Do you really want to remove this property from all selected objects?"):
					removePropFromSelected(self.name)

			buildFocusTree(g_mode)

		elif event == self.macroKeyEvent:
			self.macro_index = self.obj_mn_macro_index.val
			self.value_string = macros[self.macro_index][0]
			self.value_macro = macros[self.macro_index][1]

			for ob in self.objs:
				ob.getProperty(self.name).data = self.value_string

		elif event == self.showMacroEvent:
			self.show_macro = self.obj_tg_show_macro.val

		elif event == self.moveEvent:
			if self.type == 0:
				value = self.value_bool
				typ  = 'BOOL'
			elif self.type == 1:
				value = self.value_int
				typ = 'INT'
			elif self.type == 2:
				value = self.value_float
				typ = 'FLOAT'
			elif self.type == 3:
				value = self.value_string
				typ = 'STRING'


			if qualifs & Window.Qual.CTRL:
				if queryAction(self.moveEvent, ["Are you sure you want to remove this property from all members of this object group", "and copy it to their immediate children?"]):
					demoteProps(self.name, value, typ, self.objs)
			else:
				if queryAction(self.moveEvent, ["Are you sure you want to remove this property from all members of this object group", "and copy it to their parent object(s)?"]):
					promoteProps(self.name, value, typ, self.objs)

			buildFocusTree(g_mode)

		elif event == self.selectEvent:

			g_mousexy = Window.GetMouseCoords()
			scn = Blender.Scene.GetCurrent()

			if qualifs & Window.Qual.SHIFT and self.my_list:

				for ogo in self.my_list.objgroup_objects:
					if fallsIn(0, g_click_sel_prev[1] - g_mousexy[1], ogo.loc_y - self.loc_y):
						ogo.selectAllObjs()
				self.selectAllObjs()

			elif self.selected:
				if qualifs & Window.Qual.CTRL:
					if not self.active:
						self.active = 1
						scn.objects.active = self.objs[0]	# Make the first object in this group the active object
						g_ob_active = self.objs[0]
					else:
						#scn.objects.active.select(False)
						self.deselectAllObjs()
						self.selected = 0
				else:
					self.selected = 0
					self.deselectAllObjs()

			else:
				if qualifs & Window.Qual.CTRL:
					self.selected = 1

					self.selectAllObjs()
				else:
					scn.objects.selected = self.objs
					deselectObjGroupObjects()
					self.selected = 1

				scn.objects.active = self.objs[0]
				self.active = 1
				g_ob_active = self.objs[0]

			self.my_list.refreshSelectionStatus()
			g_click_sel_prev = g_mousexy
			Window.RedrawAll()
		Redraw()

	def selectAllObjs(self):
		for ob in self.objs:
			ob.select(1)
		self.selected = 1

	def selectAllChildren(self):
		all = [ob for ob in Blender.Scene.GetCurrent().objects if ob.type in ['Mesh', 'Empty', 'Armature', 'Lamp', 'Curve']]
		for ob in self.objs:
			ch = getGrandChildren(ob, all)
			for c in ch:
				c.select(1)

	def deselectAllObjs(self):
		for ob in self.objs:
			ob.select(0)
		self.selected = 0

	def draw(self, offsetY, maxY):

		global g_GUI_changed
		self.offset_y = offsetY
		dy = self.offset_y + self.loc_y
		if dy > maxY:
			return self.offset_y

		if self.show_macro and self.value_macro:
			lines = self.value_macro.splitlines()
			BGL.glColor3f(*CL_MACRO)
			BGL.glRecti(self.loc_x + 1  , dy + 21, self.loc_x + 804, dy + 1  - 22 * (len(lines)-1))
			BGL.glColor3f(*CL_TEXT_MD)

			for i, l in enumerate(lines):
				BGL.glRasterPos2i( self.loc_x + 12, self.loc_y + 7 +self.offset_y)
				self.offset_y -= 22
				Text(l, 'small')

		if any([ob for ob in self.objs if ob == g_ob_active]):
			self.active = 1
		else:
			self.active = 0

		if qualifs & Window.Qual.SHIFT:
			butselname = "+ Sel +"
			tipsel = "Rubber band selection of object group objects."

		elif self.selected:
			butselname = "Desel"
			tipsel = "Deselect same-depth objects with this property."
			if qualifs & Window.Qual.CTRL:
				butselname = "Act"
				tipsel = "Make the first object in this object group the active object."
				if self.active:
					butselname = "Desel"
					tipsel = "Deselect same-depth objects with this property."

		else:
			if qualifs & Window.Qual.CTRL:
				butselname = "+ Sel   "
				tipsel = "Add this object group's objects to the current selection."
			else:
				butselname = "  Sel  "
				tipsel = "Select objects in this object group."

		butname = "Delete"
		tipbut = "Remove this property from all objects in this object group."
		movebutname = "/\\"
		tipmove = "Shift this property up a level in the hierarchy, if one exists."

		if qualifs & Window.Qual.CTRL:
			butname = "Copy"
			tipbut = "Copy this property to all objects in the selected object groups."
			movebutname = "\\/"
			tipmove = "Shift this property down a level in the hierarchy, if one exists."


		dy = self.offset_y + self.loc_y
		self.obj_bt_delete				= PushButton(	butname,			self.deleteEvent,		self.loc_x,		  dy,	60, 22,														tipbut)
		self.obj_bt_select				= PushButton(	butselname,			self.selectEvent,		self.loc_x + 120, dy,	60, 22,														tipsel)
		self.obj_bt_move				= PushButton(	movebutname,		self.moveEvent,			self.loc_x + 60,  dy,	30, 22,														tipmove)

		menutypename = "Types %t|Bool %x0|Int %x1|Float %x2|String %x3"
		self.obj_str_name				= String(		"Name: ",			self.nameEvent,			self.loc_x + 180, dy,	200, 22,	self.name,				31,						"")
		if self.type == 0:
			self.obj_value_bool			= Toggle(		"True",				self.valueEvent,		self.loc_x + 380, dy,	212, 22,	self.value_bool,								"")
			self.obj_value_notbool		= Toggle(		"False",			self.value2Event,		self.loc_x + 592, dy,	212, 22,	not self.value_bool,							"")
		if self.type == 1:
			self.obj_value_int			= Number(		"",					self.valueEvent,		self.loc_x + 380, dy,	424, 22,	self.value_int,			-100000,	100000,		"")
		elif self.type == 2:
			self.obj_value_float		= Number(		"",					self.valueEvent,		self.loc_x + 380, dy,	424, 22,	self.value_float,		-100000.0,	100000.0,	"")
		elif self.type == 3:
			if self.prop_has_macro:
				self.obj_tg_show_macro	= Toggle(		"Expand Macro",		self.showMacroEvent,	self.loc_x + 380, dy,	100, 22,	self.show_macro,								"")
				self.obj_mn_macro_index	= Menu(self.macro_menu_string,		self.macroKeyEvent,		self.loc_x + 480, dy,	324, 22,	self.macro_index,								"")
			else:
				self.obj_value_string	= String(		"",					self.valueEvent,		self.loc_x + 380, dy,	424, 22,	self.value_string,		399,					"")


		BGL.glColor3f(*CL_GREY70)
		BGL.glRecti(self.loc_x + 90, dy +1, self.loc_x + 120, dy + 22)
		count_str = str(self.getObjectCount())

		if self.selected:
			if self.active:
				BGL.glColor3f(*CL_ACT_PINK)
				BGL.glRecti(self.loc_x + 92, dy +1, self.loc_x + 120, dy + 22)
			else:
				BGL.glColor3f(*CL_SEL_PINK)
				BGL.glRecti(self.loc_x + 92, dy +1, self.loc_x + 120, dy + 22)
			BGL.glColor3f(*CL_TEXT_DK)
			BGL.glRasterPos2i( self.loc_x + 116 - GetStringWidth(count_str), dy + 7 )
			Text(count_str, 'normal')
		else:
			BGL.glColor3f(*CL_GREY70)
			BGL.glRecti(self.loc_x + 92, dy +1, self.loc_x + 120, dy + 22)
			BGL.glColor3f(*CL_TEXT_MD)
			BGL.glRasterPos2i( self.loc_x + 116 - GetStringWidth(count_str), dy + 7 )
			Text(count_str, 'normal')

		return self.offset_y

# ___________________________________________________________________________________________________________________________________________________________________________________________

class ObjGroupList:


	def __init__(self):
		self.ID = 0
		self.name = ""
		self.depth = 0
		self.objgroup_objects = []
		self.ogos_sorted = []
		self.objgroup_count = 0
		self.total_obj_count = 0
		self.loc_x = 0
		self.loc_y = 0
		self.height = 0
		self.obj_bt_select_all = None
		self.obj_tg_fold	   = None
		self.select_all_event = 0
		self.fold_event = 0
		self.all_selected = 0
		self.folded = 0

	def setData(self, mode, index, depth, x, y, propname, objects):

		self.ID = index
		self.name = propname
		self.depth = depth
		self.loc_x = x
		self.loc_y = y - 22
		self.objgroup_objects = []		# a list of ObjGroupObjs
		self.objgroup_count = 0
		self.total_obj_count = 0

		vals = []						# a list of property unique values

		if objects and mode == 1:
			for ob in objects:
				for p in ob.getAllProperties():
					if stringsMatch(self.name, p.name, G_CASE_SENSITIVE):

						dat = cleanString(p.data) if p.type == 'STRING' and G_CLEANSTRINGS else p.data

						if dat in vals and self.objgroup_objects[vals.index(dat)].name == p.name and self.objgroup_objects[vals.index(dat)].selected == ob.sel:
							self.objgroup_objects[vals.index(dat)].addObject(ob)
						else:
							ogo = ObjGroupObject()
							ogo.setData(index + self.objgroup_count + 1, p.name, p.type, dat,  (index + self.objgroup_count + 1) * 20, [ob])
							ogo.selected = ob.sel
							self.objgroup_count += 1
							ogo.loc_x = self.loc_x
							ogo.loc_y = self.loc_y - 22 * self.objgroup_count
							vals.append(dat)
							self.objgroup_objects.append(ogo)

		elif objects and mode == 2:
			for ob in objects:
				for p in ob.getAllProperties():

					# filter for all props which are EFFECTIVELY the same - but maybe have different cases or value formats:

					if stringsMatch(g_prop_focus.name, p.name, G_CASE_SENSITIVE) and multiStringsMatch(g_prop_focus.value[g_prop_focus.type_enums[g_prop_focus.type]], p.data, G_CASE_SENSITIVE, G_MATCH_BY_VALUE):
						for ogo in self.objgroup_objects:

							# filter for EXACT likenesses: each OGO consists only of objects with exact likenesses!

							if ogo.selected == ob.sel and ogo.name == p.name and ogo.value[ogo.type_enums[ogo.type]] == p.data:

								# if we already have an OGO with the exact name and value as this property - add the new ob to it
								ogo.addObject(ob)
								break
						else:
							new_ogo = ObjGroupObject()
							new_ogo.setData2(index + self.objgroup_count + 1, p, (index + self.objgroup_count + 1) * 20, ob)
							new_ogo.selected = ob.sel
							self.objgroup_count += 1
							new_ogo.loc_x = self.loc_x
							new_ogo.loc_y = self.loc_y - 22 * self.objgroup_count
							self.objgroup_objects.append(new_ogo)

		elif mode == 3:		# search mode. This time we do all the string matching goo HERE

			for ob in [ob for ob in objects if ob.getAllProperties()]:

				props = getCloseProps(propname, object_mn_search_scope.val, object_mn_match_type.val, ob)
				if props:
					for p in props:
						if not self.objgroup_objects:
							new_ogo = ObjGroupObject()
							new_ogo.setData2(index + 1, p, (index + 1) * 20, ob)
							# copy the objects selection state too:
							new_ogo.selected = ob.sel
							self.objgroup_count += 1
							new_ogo.loc_x = self.loc_x
							new_ogo.loc_y = self.loc_y - 22 * self.objgroup_count
							self.objgroup_objects.append(new_ogo)
						else:
							for ogo in self.objgroup_objects:
								if ogo.matchesProp(p) and ogo.selected == ob.sel:
									ogo.addObject(ob)
									break
							else:
								o = ObjGroupObject()
								o.setData2(index + self.objgroup_count + 1, p, (index + self.objgroup_count + 1) * 20, ob)
								o.selected = ob.sel
								self.objgroup_count += 1
								o.loc_x = self.loc_x
								o.loc_y = self.loc_y - 22 * self.objgroup_count
								self.objgroup_objects.append(o)

		self.objgroup_count = len(self.objgroup_objects)

		for objgrp in self.objgroup_objects:
			self.total_obj_count += objgrp.obj_count
			objgrp.my_list = self

		self.sortGroups(0)
		self.select_all_event	= index * 20
		self.fold_event = index * 20 + 1
		self.obj_bt_select_all	= Create(0)
		self.obj_tg_fold		= Create(self.folded)

		# get the selection status of ALL the objects in ALL the object group objects in this list:
		if self.objgroup_objects:
			if all([ogo.selected for ogo in self.objgroup_objects]):
				self.all_selected = 1
		else:
			self.all_selected = 0

	def sortGroups(self, sortMode):
		""" This sorts the objgroup objects alphanumerically by property value.
			They can be sorted using any one of these four terms:
			0 - the entire prop value
			1 - the first term following the last '/'
			2 - the second term following the last '/'
			3 - the third term following the last '/'
			This allows autogen building objs to be sorted by width, depth or height by means of their 'METRICS' property, assuming they have one.
		"""
		if self.objgroup_objects:
			offsets = [o.loc_y for o in self.objgroup_objects]
			self.ogos_sorted = sorted( self.objgroup_objects, key=lambda v: getSortKey(v.value_any))
			for i, og in enumerate(self.ogos_sorted):
				og.loc_y = offsets[i]

	def refreshSelectionStatus(self):
		if not self.objgroup_objects:
			self.all_selected = 0
			return
		self.all_selected = 1
		for objgrp in self.objgroup_objects:
			objgrp.selected = 1
			for obj in objgrp.objs:
				if obj.sel == 0:
					objgrp.selected = 0
					self.all_selected = 0

	def selectChildrenOfSelected(self):
		for o in [ogo for ogo in self.objgroup_objects if ogo.selected]:
			o.selectAllChildren()

	def replaceText(self, searchScope, replaceMode, replaceScope, searchTerm, replaceTerm):
		changed = 0
		if self.objgroup_objects:
			for ogo in [ogo for ogo in self.objgroup_objects if replaceScope or (ogo.selected and not replaceScope) ]:
				changed += ogo.updateData(searchScope, replaceMode, searchTerm, replaceTerm)
		return changed

	def handleEvent(self, event):

		global g_objgrouplists

		if event == self.select_all_event:
			scn = Blender.Scene.GetCurrent()
			sltd = list(scn.objects.selected)

			if self.all_selected:					# the select button becomes a deselect button
				for ogo in self.objgroup_objects:
					ogo.deselectAllObjs()
				self.all_selected = 0
			else:
				if qualifs & Window.Qual.SHIFT:		# we add to the existing selection:
					for ogo in self.objgroup_objects:
						ogo.selectAllObjs()

				else:								# we deselect existing selections and select our own objs
					scn.objects.selected = []
					for ogo in self.objgroup_objects:
						ogo.selectAllObjs()

				self.all_selected = 1
			for ogl in g_objgrouplists:
				ogl.refreshSelectionStatus()

			Window.RedrawAll()

		elif event == self.fold_event:
			self.folded = self.obj_tg_fold.val
			Draw()

		else:
			for ogo in self.objgroup_objects:
				ogo.handleEvent(event)

	def drawList(self, offsetY, maxY):

		max_y = maxY - 22 * self.depth
		list_offset_y = offsetY
		dy = list_offset_y + self.loc_y
		header_offset_y = min(dy, max_y)
		list_offset_y += ((header_offset_y - dy)%22)
		dy = list_offset_y + self.loc_y

		# Header background colour:
		BGL.glColor3f(*CL_GREY75)
		BGL.glRecti(self.loc_x + 1, header_offset_y + 1, self.loc_x + 804, header_offset_y + 22)

		# Obj-count display background colour:
		if any([ogo.selected for ogo in self.objgroup_objects]):
			if any([ogo.active for ogo in self.objgroup_objects]):
				BGL.glColor3f(*CL_ACT_PINK)
				BGL.glRecti(self.loc_x + 92 , header_offset_y + 1, self.loc_x + 120, header_offset_y + 22)
			else:
				BGL.glColor3f(*CL_SEL_PINK)
				BGL.glRecti(self.loc_x + 92 , header_offset_y + 1, self.loc_x + 120, header_offset_y + 22)
			if not all([ogo.selected for ogo in self.objgroup_objects]):
				BGL.glColor3f(*CL_GREY75)
				BGL.glRecti(self.loc_x + 94 , header_offset_y + 3, self.loc_x + 118, header_offset_y + 20)
		else:
			BGL.glColor3f(*CL_GREY70)
			BGL.glRecti(self.loc_x + 92 , header_offset_y + 1, self.loc_x + 120, header_offset_y + 22)

		BGL.glColor3f(*CL_GREY70)
		BGL.glRecti(self.loc_x + 180 , header_offset_y + 1, self.loc_x + 380, header_offset_y + 22)
		BGL.glColor3f(*CL_TEXT_MD)
		BGL.glRasterPos2i(self.loc_x + 6, header_offset_y + 6)
		Text("Depth %d   Obs"  % self.depth)
		BGL.glRasterPos2i(self.loc_x + 184, header_offset_y + 6)
		if g_mode in [1, 2]:
			Text("Name: %s" % self.name)

		BGL.glRasterPos2i( self.loc_x + 116 - GetStringWidth(str(self.total_obj_count)), header_offset_y + 6 )
		Text(str(self.total_obj_count), 'normal')

		if self.objgroup_count:
			if self.all_selected:
				selbut_name = "Desel"
			else:
				if qualifs & Window.Qual.SHIFT:
					selbut_name = "+ Sel   "
				else:
					selbut_name = "Sel"

			self.obj_bt_select_all	= PushButton(selbut_name,	self.select_all_event,	self.loc_x + 120, header_offset_y, 60, 22, "Select/deselect all objects in this object group list.")
			self.obj_tg_fold		=	  Toggle('',            self.fold_event,		self.loc_x + 386, header_offset_y + 6, 10, 10, self.folded, "Toggle visibility of this object group list.")

		if self.folded:
			return list_offset_y + self.objgroup_count * 22
		else:
			for po in self.ogos_sorted:
				list_offset_y = po.draw(list_offset_y, header_offset_y - 22)
			return list_offset_y

# ___________________________________________________________________________________________________________________________________________________________________________________________

def saveUndoVal(name, value):
	pdict = Registry.GetKey('XPPropertyEditor', True)
	pdict[name] = value()
	Registry.SetKey('XPPropertyEditor', pdict, True)

def savePrefs():
	global G_CLEANSTRINGS, G_NO_CONFIRM, G_CASE_SENSITIVE, G_HAS_UNDO, G_AUTO_UPDATE, G_SEARCH_SCOPE, G_MATCH_TYPE, G_REPLACE_MODE, G_REPLACE_SCOPE, g_scope
	d = {}
	d['G_CLEANSTRINGS'] 	= G_CLEANSTRINGS
	d['G_NO_CONFIRM']		= G_NO_CONFIRM
	d['G_CASE_SENSITIVE']	= G_CASE_SENSITIVE
	d['G_HAS_UNDO']			= G_HAS_UNDO
	d['G_AUTO_UPDATE']		= G_AUTO_UPDATE
	d['G_SEARCH_SCOPE']		= G_SEARCH_SCOPE
	d['G_MATCH_TYPE']		= G_MATCH_TYPE
	d['G_REPLACE_MODE']		= G_REPLACE_MODE
	d['G_REPLACE_SCOPE']	= G_REPLACE_SCOPE
	d['g_scope']			= g_scope


	Registry.SetKey('XPPropertyEditor', d, True)
	pdict = Registry.GetKey('XPPropertyEditor', True)
# prefs default values:
G_MATCH_BY_VALUE		= 1
G_CLEANSTRINGS			= 1
G_NO_CONFIRM			= 0
G_CASE_SENSITIVE		= 0
G_HAS_UNDO				= 0
G_AUTO_UPDATE			= 1
G_SEARCH_SCOPE			= 0
G_MATCH_TYPE			= 1
G_REPLACE_MODE			= 0
G_REPLACE_SCOPE			= 0
G_AUTO_UPDATE			= 0



def loadPrefs():
	global G_CLEANSTRINGS, G_NO_CONFIRM, G_CASE_SENSITIVE, G_HAS_UNDO, G_AUTO_UPDATE, G_SEARCH_SCOPE, G_MATCH_TYPE, G_REPLACE_MODE, G_REPLACE_SCOPE, g_scope
	pdict = Registry.GetKey('XPPropertyEditor', True)
	if pdict:

		try:
			G_CLEANSTRINGS		= pdict['G_CLEANSTRINGS']
			G_NO_CONFIRM		= pdict['G_NO_CONFIRM']
			G_CASE_SENSITIVE	= pdict['G_CASE_SENSITIVE']
			G_HAS_UNDO			= pdict['G_HAS_UNDO']
			G_AUTO_UPDATE		= pdict['G_AUTO_UPDATE']
			G_SEARCH_SCOPE		= pdict['G_SEARCH_SCOPE']
			G_MATCH_TYPE		= pdict['G_MATCH_TYPE']
			G_REPLACE_MODE		= pdict['G_REPLACE_MODE']
			G_REPLACE_SCOPE		= pdict['G_REPLACE_SCOPE']
			g_scope				= pdict['g_scope']

		except:
			savePrefs()

def setUndoData():
	global g_undo_status
	d = {}
	d['g_proplists']		= g_proplists
	d['g_objgrouplists']	= g_objgrouplists
	d['g_prop_focus']		= g_prop_focus
	d['g_mode']				= g_mode

	d['g_ob_active']		= g_ob_active
	d['g_search_term']		= g_search_term
	d['g_replace_term']		= g_replace_term

	Blender.Registry.SetKey('XP Property Editor', d, False)
	g_undo_status = 1

	rdict = Blender.Registry.GetKey('XP Property Editor', False)

def getUndoData():
	global g_proplists, g_objgrouplists, g_prop_focus, g_mode, g_scope, g_ob_active, g_search_term, g_replace_term
	global g_undo_status
	Blender.Registry.rdict = GetKey('XP Property Editor', False)

	if rdict:
		try:
			g_proplists			= rdict['g_proplists']
			g_objgrouplists		= rdict['g_objgrouplists']
			g_prop_focus		= rdict['g_prop_focus']
			g_mode				= rdict['g_mode']

			g_ob_active			= rdict['g_ob_active']
			g_search_term		= rdict['g_search_term']
			g_replace_term		= rdict['g_replace_term']
			g_undo_status = 0
		except:
			setUndoData()

def swapUndoData():
	""" For "redo" functionality. The cached data is restored from the registry and replaced with the current data by holding it temporarily in list 'tempdata'
	"""

	global g_proplists, g_objgrouplists, g_prop_focus, g_mode, g_scope, g_ob_active, g_search_term, g_replace_term
	global g_undo_status
	vs = [ g_proplists, g_objgrouplists, g_prop_focus, g_mode, g_scope, g_ob_active, g_search_term, g_replace_term]
	ks = ['g_proplists', 'g_objgrouplists', 'g_prop_focus', 'g_mode', 'g_scope', 'g_ob_active', 'g_search_term', 'g_replace_term']
	rdict = Blender.Registry.GetKey('XPlaneBlndrPropsEdit', False)

	if rdict:
		try:

			tempdata.fromkeys(rdict.keys())
			tempdata.update(rdict)

			setUndoData()


			# Restore from tempdata:
			for v, k in zip(vs, ks):
				v = copy.copy(tempdata[k])


			# nuke the temp dict:
			tempdata = {}
			if g_undo_status == 1:
				g_undo_status = 2
			elif g_undo_status == 2:
				g_undo_status = 1
		except:
			setUndoData()

# GLOBALS ____________________________________________________________________________________________________________________________________________________________________________________________________________________

sce = Blender.Scene.GetCurrent()
g_ob_active = sce.objects.active
g_prop_focus = PropObject()
g_mode		=		0			# 0 -> normal, 1 -> name focus, 2 -> value focus, 3 -> search and replace
g_scope					= 0						# 0 -> branch, 1 -> family, 2 -> selected, 3 -> scene
g_scroll_offset = 0
g_results_text	= []
loadPrefs()

g_scope_vals = [0, 0, 0 ,0]
g_scope_vals[g_scope] = 1
object_bt_proceed		= Create(0)
object_bt_test			= Create(0)
object_bt_undo			= Create(0)
object_sel_chldrn		= Create(0)
object_tg_branch		= Create(g_scope==0)
object_tg_scene			= Create(g_scope==3)
object_st_search_term	= Create("")
object_tg_tree			= Create(g_scope==1)
object_tg_selected		= Create(g_scope==2)
object_st_replace_term	= Create("")
object_bt_search_for	= Create(0)
object_bt_close_results = Create(0)
object_cp_colours		= Create(0, 0, 0)
tg_G_MATCH_BY_VALUE		= Create(G_MATCH_BY_VALUE)	# if true, values are compared by numeric value rather than lexicographic value
tg_G_CLEANSTRINGS		= Create(G_CLEANSTRINGS)	# if true, all string-type property _values_ are sanitized using cleanString() when used as sort terms
tg_G_CONFIRMATIONS		= Create(1 - G_NO_CONFIRM)		# if true, mutes user-confirmation dialogue popups
tg_G_CASE_SENSITIVE		= Create(G_CASE_SENSITIVE)
tg_G_HAS_UNDO			= Create(G_HAS_UNDO)		# if true, provide undo functionality
tg_G_AUTO_UPDATE		= Create(G_AUTO_UPDATE)		# If true, lists update as soon as mouse enters window.
object_mn_search_scope  = Create(G_SEARCH_SCOPE)	#  0: names, 1: values
object_mn_match_type	= Create(G_MATCH_TYPE)		#  0: which match, 1: which contain
object_mn_replace_mode	= Create(G_REPLACE_MODE)	#  0: substring , 1: entire string
object_mn_replace_scope = Create(G_REPLACE_SCOPE)	#  0: in selected matches, 1: in all matches
object_mn_sort_by		= Create(1)
object_bt_sort_by_w		= Create(0)
object_bt_sort_by_d		= Create(0)
object_bt_sort_by_h		= Create(0)
object_bt_yes			= Create(0)
object_bt_no			= Create(0)
g_search_term			= ""
g_replace_term			= ""

# ____________________________________________________________________________________________________________________________________________________________________________________________________________________________

def waitConfirm(msg):
	global state
	showResults(msg)
	state = 0
	id = Window.GetAreaID()

	while not state:
		e, v = Window.QRead()
		Window.QHandle(id)
		if e:
			Window.QAdd(id, e, v)
		if e == XKEY:
			state = 1
			break
		if state:

			break
	print "state:", state
	return state

def cleanUp():
	#if [tx for tx in Blender.Text.Get() if tx.name == g_shtext.name]:
	#Blender.Text.unlink(g_shtext)
	savePrefs()

def smartReplace(string, old, new, caseSensitive=0):
	"""
	This is smarter than the standard python function replace().
	If caseSensitive is set to zero, the function will do a non-case-sensitive _search_,
	but a case sensitive _replace_...
	i.e. the case of both the replace string and the non-replaced parts of the original string will remain unchanged.
	"""
	if caseSensitive or old == "":
		return string.replace(old, new)
	else:
		i = 0
		j = 0
		d = len(old)
		k = 0
		outstr = ""
		while j > -1 and k < 40:
			j = string.upper().find(old.upper(),i)
			if  j == -1:
				outstr += string[i:]
				return outstr
			else:
				outstr += string[i:j]
				outstr += new
				i = j + len(old)
				k += 1
		return outstr

def plural(quantity, sng=None, plr=None):

	if not sng and not plr:
		return "" if quantity == 1 else "s"
	return sng if quantity == 1 else plr

def isStandardName(inStr):
	if not isinstance(inStr, str):
		return False
	if len(inStr) > 12 and inStr[0:4].isdigit() and inStr[4].isalpha() and inStr[5:9].isdigit() and inStr[9].isalpha() and inStr[10:13].isdigit():
		return True
	else:
		return False

def isNumber(inStr):
	try:
		float(inStr)
		return True
	except ValueError:
		return False

def isIntString(inStr):
	if not isNumber(inStr):
		return False
	return isinstance(eval(str(inStr)), int)	# cast to string in case input is a float or an int

def isFloatString(inStr):
	if not isNumber(inStr):
		return False
	return isinstance(eval(str(inStr)), float)	# cast to string in case input is a float or an int

def stringsMatch(str1, str2, caseSensitive=0, partialMatch=0):
	if not caseSensitive:
		if partialMatch:
			return str1.upper().strip() in str2.upper()
		else:
			return str1.upper().strip() == str2.upper().strip()
	else:
		if partialMatch:
			return str1.strip() in str2
		else:
			return str1.strip() == str2.strip()

def multiStringsMatch(inStr1, inStr2, caseSensitive=1, matchByValue=1):
	"""
	Test two single or multiple-term strings for equal value.
	Args can be strings, floats or ints.
	The following are considered to be true:
	1.0 == 1
	+1 == 1
	.1 == 0.1
	'Word' == 'word' if caseSensitive == 0, else false.

	Note: this function is case sensitive by default.
	"""
	tks1 = str(inStr1).split()
	tks2 = str(inStr2).split()
	# test for equal number of terms:
	if len(tks1) != len(tks2):
		return False

	for a, b in zip(tks1, tks2):

		if matchByValue:

			if isNumber(a) and isNumber(b):
				if eval(a) == eval(b):
					continue
				else:
					return False

			elif not isNumber(a) and not isNumber(b):
				if not caseSensitive:
					if a.upper() == b.upper():
						continue
					else:
						return False
				else:
					if a == b:
						continue
					else:
						return False
			else:
				return False
		else:

			if not caseSensitive:
				if a.upper() == b.upper():
					continue
				else:
					return False
			else:
				if a == b:
					continue
				else:
					return False

	return True

def pointInRect(point, rect):

	""" takes a list of two floats and a list of four floats.
		Returns True if the point [x, y] is inside rectangle rect [x1, y1, x2, y2]
	"""
	if len(rect) != 4 or len(point) != 2:
		return False
	return rect[0] <= point[0] <= rect[2] and rect[1] <= point[1] <= rect[3]

def fallsIn(limit1, limit2, x):

	return limit1 <= x <= limit2 if limit2 > limit1 else limit1 >= x >= limit2

def cleanString(inValue):

	"""
	This strips whitespace and commas, and adds two-space separators between tokens
	"""
	return "  ".join((" ".join(str(inValue).split(','))).split())

def getSortKey(inValue):

	if G_CLEANSTRINGS:
		v = cleanString(inValue)
	else:
		v = str(inValue)
	if not v:
		return v
	if g_sort_mode == 0:
		return	v
	else:
		leaf = v.rpartition("/")[2]
		if isStandardName(leaf):
			splits = [leaf[0:5], leaf[5:10], leaf[10:]]
			if g_sort_mode - 1 < len(splits):
				return splits[g_sort_mode - 1]
			else:
				return splits[0]
		else:
			splits = leaf.split()
			if g_sort_mode - 1 < len(splits):
				term = splits[g_sort_mode - 1]
				if not term.strip('0123456789.-+'):
					return float(term)
				else:
					return term
			else:
				return splits[0]

def deselectObjGroupObjects():
	global g_objgrouplists
	for ogl in g_objgrouplists:
		for ogo in ogl.objgroup_objects:
			ogo.selected = 0

def setActiveObj(obj):
	global g_ob_active
	g_ob_active = obj
	Blender.Scene.GetCurrent().objects.active = g_ob_active

def resetFocus(focusIndex):
	global g_proplists
	for pl in g_proplists:
		for p in pl.prop_objects:
			if p.index != focusIndex:
				p.is_focus = 0

def isGrandParent(child, grandparent):
	while child != None:
		if child == grandparent:
			return True
		child = child.getParent()
	return False

def isParent(child, parent):
	return child.getParent() == parent

def getChildren(parent, all):
	c=[]
	for o in all:
		if isParent(o,parent):
			c.append(o)
	return c

def getGrandChildren(parent, all):
	c=[]
	for o in all:
		if isGrandParent(o,parent):
			c.append(o)
	return c

def getGrandParent(obj):
	while obj.parent:
		obj = obj.parent
	return obj

def getBranch(obj):
	obs = []
	ob = obj
	while ob:
		obs.append(ob)
		ob = ob.parent
	return obs

def getFamily(obj, all):
	return  getGrandChildren(getGrandParent(obj), all)

def isDepth(child, d):
	while d > 0:
		if child == None:
			return False
		child = child.getParent()
		d = d-1
	if child == None:
		return False
	return child.getParent() == None

def getAllDepth(d, all):
	c=[]
	for o in all:
		if isDepth(o,d):
			c.append(o)
	return c

def getDepth(obj):
	d = 0
	while obj.parent:
		obj = obj.parent
		d += 1
	return d

def getSameNameProp(propname, all):
	samename = []
	for ob in all:
		for p in ob.getAllProperties():
			if stringsMatch(p.name, propname, G_CASE_SENSITIVE):
				samename.append(ob)
	return samename

def getSameNameValueProp(propname, propvalue, all):
	samenamevalue = []
	for ob in all:
		for p in ob.getAllProperties():
			if stringsMatch(p.name, propname, G_CASE_SENSITIVE) and multiStringsMatch(p.data, propvalue, G_CASE_SENSITIVE, G_MATCH_BY_VALUE):
				samenamevalue.append(ob)
	return samenamevalue

def getCloseProps(searchString, searchScope, matchType, ob):

	if not ob.getAllProperties():
		return []
	closeprops = []
	props = ob.getAllProperties()

	for p in props:
		if searchScope == 0:	# names
			if matchType == 0:		# exact match
				if stringsMatch(searchString, p.name, G_CASE_SENSITIVE):
					closeprops.append(p)
			elif matchType == 1:	# partial match
				if G_CASE_SENSITIVE:
					if searchString in p.name:
						closeprops.append(p)
				else:
					if searchString.upper() in p.name.upper():
						closeprops.append(p)

		if searchScope == 1:	# values
			if matchType == 0:		# exact match
				if multiStringsMatch(searchString, str(p.data), G_CASE_SENSITIVE, G_MATCH_BY_VALUE):
					closeprops.append(p)
			elif matchType == 1:	# partial match
				if G_CLEANSTRINGS:
					if G_CASE_SENSITIVE:
						if cleanString(searchString) in cleanString(p.data):
							closeprops.append(p)
					else:
						if cleanString(searchString).upper() in cleanString(p.data).upper():
							closeprops.append(p)
				else:
					if G_CASE_SENSITIVE:
						if searchString in str(p.data):
							closeprops.append(p)
					else:
						if searchString.upper() in str(p.data).upper():
							closeprops.append(p)
	return closeprops

def updateObjPropsList(ob, propName, propValue, propType, caseSensitive=0):
	"""
		Blender boolean-type properties change type to 'INT' when their data is changed.
		The workaround is to delete them and add a new prop with the new data.
		In order to maintain the order of props in the list, we have to replace the entire list...
	"""
	details = []
	for p in ob.getAllProperties():
		details.append([p.name, propValue, propType] if stringsMatch(p.name, propName, G_CASE_SENSITIVE) else [p.name, p.data, p.type])
	ob.removeAllProperties()
	for d in details:
		ob.addProperty(*d)

def removePropFromSelected(propName):

	global g_results_text
	obs = Blender.Scene.GetCurrent().objects.context
	if not obs:
		showResults("CAUTION: No objects are selected to operate on!")
		return
	r = 0
	for ob in obs:
		for p in ob.getAllProperties():
			if stringsMatch(propName, p.name, G_CASE_SENSITIVE):
				ob.removeProperty(p.name)
				r += 1
				break
	buildTree(1)
	Window.RedrawAll()
	results = []
	if r:
		results.append("A property was removed from %d object%s." % (r, plural(r)))
	else:
		results.append("CAUTION: No matching properties found - no actions were taken.")
	showResults(results)

def copyPropToSelected(propName, propValue, propType):

	global g_results_text
	obs = Blender.Scene.GetCurrent().objects.context
	if len(obs) < 1:
		g_results_text = []
		g_results_text.append("CAUTION: No objects are selected to copy to!")
		Draw()
		return
	m = 0
	a = 0

	for ob in  obs:
		for p in ob.getAllProperties():
			if stringsMatch(propName, p.name, G_CASE_SENSITIVE) and p.type == propType:
				if not multiStringsMatch(propValue, p.data, G_CASE_SENSITIVE, G_MATCH_BY_VALUE):
					if propType == 'BOOL':
						updateObjPropsList(ob, propName, propValue, propType, G_CASE_SENSITIVE)
					else:
						p.data = propValue
					m += 1
				break
		else:
			ob.addProperty(propName, propValue, propType)
			a += 1

	buildTree(1)
	Window.RedrawAll()
	u =  min(0, len(obs) - (m + a ))
	results = []
	if m:
		results.append("A property was modified on %d object%s." % (m, plural(m)))
	if a:
		results.append("A property was added to %d object%s."	% (a, plural(a)))
	if u:
		results.append("CAUTION: No changes were made to %d object%s because the active property and value were already present." % (u, plural(u)))
	showResults(results)

def promoteProps(propName, propValue, propType, objects):
	m = 0
	a = 0
	r = 0
	for ob in objects:
		if ob.parent != None:
			prnt = ob.parent
			for p in prnt.getAllProperties():
				if stringsMatch(propName, p.name, G_CASE_SENSITIVE) and p.type == propType:
					if not multiStringsMatch(propValue, p.data, G_CASE_SENSITIVE, G_MATCH_BY_VALUE):
						if propType == 'BOOL':
							updateObjPropsList(prnt, propName, propValue, propType, G_CASE_SENSITIVE)
						else:
							p.data = propValue
						m += 1
					break
			else:
				prnt.addProperty(propName, propValue, propType)
				a += 1
			r += 1
			ob.removeProperty(propName)
	results = []
	rr = ""
	if r:
		rr += "Property '%s' with value '%s' was removed from %d object%s" % (propName, str(propValue), r, plural(r))
	if a:
		rr += " and added to %d object%s." % ( a, plural(a))
	else:
		rr += "."
	results.append(rr)
	if m:
		results.append( "The value of an existing property with the same name was updated on %d object%s." % ( m, plural(m)))
	showResults(results)

def demoteProps(propName, propValue, propType, objects):
	m = 0
	a = 0
	r = 0
	for ob in objects:
		chldrn = getChildren(ob, Blender.Scene.GetCurrent().objects)
		if chldrn:
			for c in chldrn:
				if c.type in ['Mesh', 'Empty', 'Lamp', 'Curve']:
					for p in c.getAllProperties():
						if stringsMatch(propName, p.name, G_CASE_SENSITIVE) and p.type == propType:
							if propType == 'BOOL':
								updateObjPropsList(c, propName, propValue, propType, G_CASE_SENSITIVE)
							else:
								p.data = propValue
							m += 1
							break
					else:
						c.addProperty(propName, propValue, propType)
						a += 1
			ob.removeProperty(propName)
			r += 1
	results = []
	if r:
		results.append("A property was removed from %d object%s and copied to %d object%s." % (r, plural(r), a, plural(a)))
	if m:
		results.append("%s modified on %d object%s." % (plural(m,'A property was', 'Properties were'), m, plural(m)))
	showResults(results)
	Window.RedrawAll()

def createPropList(object, depth):

	global g_index
	global g_loc_x
	global g_loc_y
	new_list = ObPropList(g_index, object, depth, g_loc_x, g_loc_y)
	return new_list

def createObjGroupList(depth, propname, objects):
	global g_mode
	global g_ogindex
	global g_og_loc_x
	global g_og_loc_y
	newlist = ObjGroupList()
	newlist.setData(g_mode, g_ogindex, depth, g_og_loc_x, g_og_loc_y, propname, objects)

	return newlist

def buildTree(refresh_props_only=0):

	global g_proplists, g_ob_active, g_loc_x, g_loc_y, g_index

	width, height = Window.GetAreaSize()
	obs = []

	if refresh_props_only:

		# We want to keep the current obj tree, just reload the props and rebuild the display.
		for pl in g_proplists:
			obs.append(pl.obj)

	else:
		# We want to build a new display tree, from the active object upwards.
		sce = Blender.Scene.GetCurrent()
		g_ob_active = sce.objects.active
		ob = g_ob_active
		if not ob:
			return 0
		obs.append(ob)
		while ob.parent:
			ob = ob.parent
			obs.append(ob)
			ob.getParent()
		obs.reverse()

	g_proplists = []
	g_loc_x = 10
	g_loc_y = height -  92
	g_index = 50
	depth = 0
	for o in obs:
		new_list = createPropList(o, depth)
		g_proplists.append(new_list)
		g_index += new_list.prop_count
		g_index += 1
		g_loc_y -= (22 * new_list.prop_count + 32)
		depth += 1
	Window.RedrawAll()
	return depth

def buildFocusTree(mode, refresh_props_only=0):

	global g_objgrouplists
	global g_mode, g_ogindex
	global g_scope
	global g_prop_focus
	global g_sort_mode, g_search_flag
	global sce
	global g_og_loc_x, g_og_loc_y

	width, height = Window.GetAreaSize()
	sce = Blender.Scene.GetCurrent()
	obact = sce.objects.active
	g_og_loc_x = 10
	g_og_loc_y = height - 92
	g_ogindex = 100
	g_mode = mode

	if not obact:
		showResults("ERROR: No active object!\n")
		return

	fold_status = []
	for ogl in g_objgrouplists:
		fold_status.append(ogl.folded)

	g_objgrouplists = []

	all = [ob for ob in sce.objects if ob.type in ['Mesh', 'Empty', 'Lamp', 'Armature', 'Curve']]

	obs = []
	if g_scope == 0:		# branch
		obs = getBranch(obact)
	elif g_scope == 1:		# tree
		obs = getFamily(obact, all)
	elif g_scope == 2:		# selected
		obs = [ob for ob in sce.objects.selected if ob.type in ['Mesh', 'Empty', 'Lamp', 'Armature', 'Curve']]
	elif g_scope == 3:		# scene
		obs = all

	# find the maximum depth within the specified scope:
	max_depth = 0
	for ob in obs:
		max_depth = max(max_depth, getDepth(ob))

	bin = []
	dep = 0

	if not g_prop_focus:
		return 0

	# special case properties with name 'METRICS' - provide options for sorting the display list
	sortable = ['METRICS', 'VNAME', 'RNAME', 'EXTERNAL']

	tok = g_prop_focus.name.upper()[0:8]

	for tk in sortable:
		if tk in tok:
			if not g_sort_mode:
				g_sort_mode = 1
			break
	else:
		g_sort_mode = 0

	if g_mode == 1:
		while dep <= max_depth and g_ogindex < 700:
			bin = getSameNameProp(g_prop_focus.name, getAllDepth(dep, obs))
			newlist = createObjGroupList(dep, g_prop_focus.name, bin)
			g_objgrouplists.append(newlist)
			g_og_loc_y -= (22 * newlist.objgroup_count + 32)
			g_ogindex += newlist.objgroup_count
			g_ogindex += 1

			dep += 1

	elif g_mode == 2:
		if g_prop_focus.type == 0:
			propval	= g_prop_focus.value_bool
		elif g_prop_focus.type == 1:
			propval	= g_prop_focus.value_int
		elif g_prop_focus.type == 2:
			propval	= g_prop_focus.value_float
		elif g_prop_focus.type == 3:
			propval	= g_prop_focus.value_string

		while  dep <= max_depth and g_ogindex < 700:
			bin = getSameNameValueProp(g_prop_focus.name, propval, getAllDepth(dep, obs))

			newlist = createObjGroupList(dep, g_prop_focus.name, bin)
			g_objgrouplists.append(newlist)
			g_og_loc_y -= (22 * newlist.objgroup_count + 32)
			g_ogindex += newlist.objgroup_count
			g_ogindex += 1

			dep += 1

	elif g_mode == 3 and g_search_flag == 1:

		while dep <= max_depth and g_ogindex < 700:
			bin = getAllDepth(dep, obs)  # this time we'll do all the string matching in the objgrouplist

			newlist = createObjGroupList(dep, g_prop_focus.name, bin)
			g_objgrouplists.append(newlist)
			g_og_loc_y -= (22 * newlist.objgroup_count + 32)
			g_ogindex += newlist.objgroup_count
			g_ogindex += 1
			dep += 1

	# restore the previous fold status of each list:
	for f, ogl in zip(fold_status, g_objgrouplists):
		ogl.folded = f

	if g_ogindex >= 600:
		g_objgrouplists = []
		showResults("ERROR: Too many results! Try a more specific search.")
	Draw()

	return dep

def doSearch( searchScope, matchType, searchString, objScope):

	global g_prop_focus, g_mode, g_search_flag

	# reinit the global prop object
	if searchString:
		g_prop_focus.name = searchString
		g_prop_focus.type = 3
		g_prop_focus.value_any = searchString
		g_prop_focus.value_string = searchString
		g_search_flag = 1
		g_mode = 3
		buildFocusTree(3)

def doReplace2(searchScope, replaceMode, replaceScope, searchTerm, replaceTerm):

	global g_objgrouplists
#	setUndoData()
	changed = 0
	for ogl in g_objgrouplists:
		changed += ogl.replaceText(searchScope, replaceMode, replaceScope, searchTerm, replaceTerm)
	return changed

def queryAction(event, msg=None):

	global g_event_queue, g_confirmed, g_results_text, qualifs

	if not isinstance(event, int):
		showResults("ERROR: Non-integer value passed to queryAction()!")
		return 0


	#If confirmations are turned off in prefs, approve the action by returning 1:

	if G_NO_CONFIRM:
		return 1


	elif g_confirmed:
	# If the event has been confirmed, reset the queue and snd set g_confirmed back to zero ready for the next event. Then return 1 to action the event.
	# The cached key qualifier state has already done its job of routing the confirmation event back to us, so we can now reset it.

		g_event_queue = [0, 0]			# event code, key qualifier state
		g_confirmed = 0
		qualifs = Window.GetKeyQualifiers()

		return 1

	else:
	# If the event has not been confirmed it means it is our first call for this action.
	# Queue the event code and open a dialog in the gui.
	# Giving a value to g_event_queue will divert all subsequent incoming button events to #23 (confirm) and 24 (cancel)
	# Note that we also have to cache the key qualifiers at the state they were in at the time of the original call

		qualifs =  Window.GetKeyQualifiers()
		g_event_queue = [event, qualifs]
		if msg:
			if isinstance(msg, list):
				g_results_text = msg
			elif isinstance(msg, str):
				g_results_text = [msg]
			else:
				g_results_text = ["Are you sure? "]
		else:
			g_results_text = ["Are you sure? "]
		gui()
		Draw()
		return 0

def showResults(msg):

	global g_results_text
	if isinstance(msg, list):
		g_results_text = msg
	elif isinstance(msg, str):
		g_results_text = [msg]
	if not g_results_text:
		g_results_text.append("ERROR: It was not possible to perform the requested operation.")
	gui()
	Window.RedrawAll()

def event(evt, val):

	global qualifs
	global g_scroll_offset
	global g_ob_active, g_my_window_cds, g_mousexy, g_results_text


	if evt == ESCKEY:
		rep = PupMenu(" REALLY QUIT?:%t|    NO|  YES")
		if rep == 2:
			cleanUp()
			Exit()
			return

	elif evt == PAGEUPKEY and val:
		g_scroll_offset = max(0, g_scroll_offset - 16)
		gui()
		Draw()
	elif evt == PAGEDOWNKEY and val:
		g_scroll_offset += 16
		gui()
		Draw()

	elif evt == HOMEKEY and val:
		g_scroll_offset = 0
		gui()
		Draw()

	elif evt == WHEELUPMOUSE and val:
		g_scroll_offset = max(0, g_scroll_offset - 1)
		gui()
		Draw()

	elif evt == WHEELDOWNMOUSE and val:
		g_scroll_offset += 1
		gui()
		Draw()

	elif evt in [LEFTSHIFTKEY, RIGHTSHIFTKEY,	LEFTCTRLKEY, RIGHTCTRLKEY]:
		qualifs = Window.GetKeyQualifiers()
		Draw()

	elif evt in [MOUSEX, MOUSEY]:
		if G_AUTO_UPDATE:
			x, y = list(Window.GetMouseCoords())
			if abs(x- g_mousexy[0]) > 25 or abs(y - g_mousexy[1]) > 25:
#			print Window.GetMouseCoords()
				if	Blender.Scene.GetCurrent().objects.active:
					ob_act = Blender.Scene.GetCurrent().objects.active
					if g_ob_active != ob_act:
						g_ob_active = ob_act
						if g_mode == 0:
							buildTree()
							Draw()
						else:
							buildFocusTree(g_mode)
							Draw()
			g_mousexy = [x, y]

def button_event(evt):

	global g_proplists, g_objgrouplists
	global g_loc_x
	global g_loc_y
	global g_index
	global g_mode, g_scope, g_search_term, g_replace_term, g_sort_mode
	global G_NO_CONFIRM, G_HAS_UNDO, G_CASE_SENSITIVE, G_CLEANSTRINGS, G_AUTO_UPDATE, G_SEARCH_SCOPE, G_MATCH_TYPE, G_REPLACE_MODE, G_REPLACE_SCOPE, G_MATCH_BY_VALUE
	global sce
	global object_bt_test, object_tg_branch, object_tg_tree, object_tg_selected, object_tg_scene, object_st_search_term, object_st_replace_term, object_bt_prefs
	global object_mn_search_scope, object_mn_match_type,  object_mn_replace_mode, object_mn_replace_scope, object_bt_close_results
	global g_scope_vals
	global g_scroll_offset
	global g_results_text
	global g_event_queue, g_confirmed, qualifs
	global proceed

	if g_results_text:
		g_results_text = []

	if evt == 25:
		proceed = 1

	# wait for user
	if g_event_queue and not g_confirmed:

		if evt == 23:

			# Confirm
			g_results_text = []
			g_confirmed = 1
			qualifs = g_event_queue[1]
			evt = g_event_queue[0]
			state = 1
			# allow the approved event code to trickle down through the if block to the correct recipient:

		elif evt == 24:
			state = -1
			# Cancel. Reset everything and prevent the event from reaching the action handler:
			g_results_text = []
			g_confirmed = 0
			g_event_queue = [0, 0]
			showResults("CANCELLED.")
			gui()
			Draw()
			id = Window.GetAreaID()
			t0 = sys.time()
			while sys.time() - t0 < 0.5:
				e, v = Window.QRead()
				if e not in [MOUSEX, MOUSEY] and v:
					break
				Window.QAdd(id, e, v)
			g_results_text = []

			Draw()
			return

	if evt == 4:
		# select children of selected
		if g_mode:
			for ogl in g_objgrouplists:
				ogl.selectChildrenOfSelected()
		else:
			for po in g_proplists:
				po.selectChildren()
		Window.RedrawAll()

	elif evt == 5:
		# scope: branch
		if g_scope == 0:
			Draw()
			return
		g_scope = 0
		g_scope_vals[0] = 1
		g_scope_vals[1] = g_scope_vals[2] = g_scope_vals[3] = 0
		if g_mode == 3:
			g_search_flag = 1
		buildFocusTree(g_mode)

	elif evt == 6:
		# scope: tree
		if g_scope == 1:
			Draw()
			return
		g_scope = 1
		g_scope_vals[1] = 1
		g_scope_vals[0] = g_scope_vals[2] = g_scope_vals[3] = 0
		if g_mode == 3:
			g_search_flag = 1
		buildFocusTree(g_mode)

	elif evt == 7:
		#scope: selected
		if g_scope == 2:
			Draw()
			return
		g_scope = 2
		g_scope_vals[2] = 1
		g_scope_vals[0] = g_scope_vals[1] = g_scope_vals[3] = 0
		if g_mode == 3:
			g_search_flag = 1
		buildFocusTree(g_mode)

	elif evt == 8:
		#scope: scene
		if g_scope == 3:
			Draw()
			return
		g_scope = 3
		g_scope_vals[3] = 1
		g_scope_vals[0] = g_scope_vals[1] = g_scope_vals[2] = 0
		if g_mode == 3:
			g_search_flag = 1
		buildFocusTree(g_mode)

	elif evt == 9:
		# Prefs popup dialogue

		block = []

		block.append(("Confirmations", tg_G_CONFIRMATIONS, "Turns off action-confirmation dialogues."))
	#	block.append(("Allow undo", tg_G_HAS_UNDO, ""))
		block.append(("Case Sensitive", tg_G_CASE_SENSITIVE, "Turns on case sensitivity for searches."))
		block.append(("Relaxed Search", tg_G_CLEANSTRINGS, "Searches are performed using sanitized versions of string values"))
		block.append(("Match by Value", tg_G_MATCH_BY_VALUE, "Property values are matched by numeric value e.g:'1' = '1.0'"))
		block.append(("Auto Refresh", tg_G_AUTO_UPDATE, "The display updates as soon as the selection changes."))
		retval = PupBlock("Preferences", block)
		if retval:
			G_MATCH_BY_VALUE    = tg_G_MATCH_BY_VALUE.val
			G_CLEANSTRINGS		= tg_G_CLEANSTRINGS.val
			G_NO_CONFIRM		= 1 - tg_G_CONFIRMATIONS.val
			g_confirmed			= G_NO_CONFIRM
			G_CASE_SENSITIVE	= tg_G_CASE_SENSITIVE.val
		#	G_HAS_UNDO			= tg_G_HAS_UNDO.val
			G_AUTO_UPDATE		= tg_G_AUTO_UPDATE.val
		showResults(["* Confirmation dialogues are turned %s." % ("off. Actions will be performed immediately" if G_NO_CONFIRM else "on"),
					"* Case sensitivity is turned %s." % ("on" if G_CASE_SENSITIVE else "off"),
					"* %s attitude to whitespace is taken while performing searches." %("A relaxed" if G_CLEANSTRINGS else "An exact" ),
					"* Matching is performed by %s value, not %s value." % ("numeric" if G_MATCH_BY_VALUE else "lexicographic", "lexicographic" if G_MATCH_BY_VALUE else "numeric"),
					"* The property list displays %s updated %s to reflect changes." % ("are" if G_AUTO_UPDATE else "must be", "automatically" if G_AUTO_UPDATE else "manually")])

	elif evt == 10:
		g_search_term = object_st_search_term.val

	elif evt == 11:
		g_replace_term = object_st_replace_term.val

#	elif evt == 12:

	elif evt == 13:
		doSearch( object_mn_search_scope.val, object_mn_match_type.val,  object_st_search_term.val, g_scope)

	elif evt == 14:
		G_SEARCH_SCOPE = object_mn_search_scope.val
		doSearch( object_mn_search_scope.val, object_mn_match_type.val,  object_st_search_term.val, g_scope)

	elif evt == 15:
		G_MATCH_TYPE = object_mn_match_type.val
		doSearch( object_mn_search_scope.val, object_mn_match_type.val,  object_st_search_term.val, g_scope)

	elif evt == 16:
		if queryAction(16):
			changed = doReplace2(object_mn_search_scope.val, object_mn_replace_mode.val, object_mn_replace_scope.val, object_st_search_term.val, object_st_replace_term.val)
			if changed:
				showResults("%s%d substitution%s performed." % (plural(changed,"", "A total of "), changed, plural(changed, " was", "s were")))
			else:
				showResults("CAUTION: No action was taken.")

	elif evt == 17:
		G_REPLACE_MODE = object_mn_replace_mode.val

	elif evt == 18:
		G_REPLACE_SCOPE = object_mn_replace_scope.val

	elif evt == 19:
	# sort order menu
		g_sort_mode = object_mn_sort_by.val
		buildFocusTree(g_mode)

	elif evt == 20:
	# Close results panel
		g_results_text = []
		Draw()

	elif g_mode == 0:
		# Prop list objects
		for pl in g_proplists:
			pl.handleEvent(evt)

		if evt == 1:
		# get active
			if not sce.objects.active:
				return
			buildTree()

	elif g_mode > 0:
		# Object group objects
		for ogl in g_objgrouplists:
			ogl.handleEvent(evt)

		if evt == 1:
		# release focus
			g_mode = 0
			g_scroll_offset = 0
			buildTree(1 - G_AUTO_UPDATE)

	Draw()

def gui():

	global object_bt_test, object_tg_branch, object_tg_tree, object_tg_selected, object_tg_scene, object_bt_prefs, object_tg_case_sensitive, object_sel_chldrn, object_bt_undo
	global object_st_search_term, object_st_replace_term,	object_bt_search_for, object_bt_replace_with, object_mn_search_scope, object_mn_match_type,	object_mn_replace_mode, object_mn_replace_scope, object_mn_sort_by, object_bt_sort_by_w, object_bt_sort_by_d, object_bt_sort_by_h, object_cp_colours, object_bt_yes, object_bt_no
	global qualifs
	global g_proplists, g_objgrouplists
	global g_loc_x
	global g_loc_y
	global g_index
	global g_mode, g_search_term, g_replace_term, g_sort_mode, g_undo_status
	global g_scope_vals
	global g_scroll_offset
	global g_my_window_cds, windowtop
	global myWindowID
	global g_results_text

	BGL.glClearColor(0.60, 0.60, 0.60, 1)
	BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)

	w, h = Window.GetAreaSize()
	windowtop = h - 10

	searchscopename = "         names %x0|         values %x1"
	matchtypename =   "   exactly matching %x0|       containing %x1"
	replacemodename = "     search string  %x0|      entire string %x1"
	replacescopename = "    in selected with %x0|   in all results with %x1"

	object_bt_prefs =		PushButton("Preferences",		 9, 290, windowtop - 72,	100, 22,										"Preference settings")
	object_bt_search_for =	PushButton("Search for",		13,  10, windowtop - 22,	100, 22,										"")
#	object_bt_undo		 =	PushButton(undoname,			 3, 190, windowtop - 72,	100, 22,										"")
	object_mn_search_scope =	  Menu(searchscopename,		14,	110, windowtop - 22,	140, 22,	object_mn_search_scope.val,			"")
	object_mn_match_type =		  Menu(matchtypename,		15,	250, windowtop - 22,	140, 22,	object_mn_match_type.val,			"")
	object_bt_replace_with =PushButton("Replace",			16,  10, windowtop - 44,	100, 22,										"")
	object_mn_replace_mode =	  Menu(replacemodename,		17,	110, windowtop - 44,	140, 22,	object_mn_replace_mode.val,			"")
	object_mn_replace_scope =	  Menu(replacescopename,	18,	250, windowtop - 44,	140, 22,	object_mn_replace_scope.val,		"")
	object_st_search_term =		String("",					10,	390, windowtop - 22,	424, 22,	object_st_search_term.val,	200,	"")
	object_st_replace_term =	String("",					11,	390, windowtop - 44,	424, 22,	object_st_replace_term.val,	200,	"")
	object_tg_branch =			Toggle("Branch",			5,	502, windowtop - 72,	78,	 22,	g_scope_vals[0],					"")
	object_tg_tree =			Toggle("Tree",				6,	580, windowtop - 72,	78,	 22,	g_scope_vals[1],					"")
	object_tg_selected =		Toggle("Selected",			7,	658, windowtop - 72,	78,	 22,	g_scope_vals[2],					"")
	object_tg_scene =			Toggle("Scene",				8,	736, windowtop - 72,	78,	 22,	g_scope_vals[3],					"")
#	object_cp_colours	= ColorPicker(						21, 820, windowtop - 104,	94, 94,		object_cp_colours.val,				"")
	object_sel_chldrn	=   PushButton("Select Children",	4, 190,  windowtop - 72,	100, 22,										"Select children of selected objects.")

	BGL.glColor3f(*CL_TEXT_MD)
	BGL.glRasterPos2i( 422,  windowtop - 66 )
	Text("Focus scope", 'normal')

	# results offset: nudge the displayed lists down a bit when we show results of user actions:

	if g_results_text:
		BGL.glColor3f(0.22, 0.22, 0.22)
		BGL.glRecti(130, windowtop - 82, 814,  windowtop - 82 - len(g_results_text) * 22)
		ofst = windowtop - 98
		xloc = 140
		for l in g_results_text:
			if l.startswith("CAUTION"):
				col = CL_ORANGE
			elif g_event_queue[0] and not g_confirmed:
				xloc = 140
				object_bt_yes = PushButton("Yes",			23, 10, windowtop - 104,	60, 22,											"Apply changes")
				object_bt_no =	PushButton("No",			24, 70, windowtop - 104,	60, 22,											"Cancel")
				col = CL_ORANGE
			elif l.startswith("ERROR"):
				col = CL_RED

			else:
				col = CL_CYAN
			BGL.glColor3f(*col)
			BGL.glRasterPos2i(xloc,  ofst)
			Text(l, 'normal')
			BGL.glRasterPos2i(xloc,  ofst)
			Text(l, 'normal')  # Draw text twice because the font is rather lightweight
			ofst -= 22
		if not g_event_queue[0] or g_confirmed:
			object_bt_close_results = PushButton("Close",			20, 10, windowtop - 104,	120, 22,					"Close results panel.")

	ro = 0 if not g_results_text else -(10 + len(g_results_text) * 22)

	y_offset = ro + (22 * g_scroll_offset)

	if g_mode == 0:
		object_bt_test	=	PushButton("Get Active", 		1,	10, windowtop - 72,		180, 22,					"")
		for pl in g_proplists:
			y_offset = pl.drawList(y_offset, windowtop + ro - 104)

	elif g_mode > 0:

		object_bt_test	=	PushButton("Release Focus", 	1,	10,	 windowtop - 72,	180, 22,					"")

		for og in g_objgrouplists:
			y_offset = og.drawList(y_offset, windowtop + ro - 104)

		if g_sort_mode:
		# If the focus proerty name is 'METRICS', 'VNAME', 'RNAME' or 'external' we draw a menu giving options for sorting the list of metrics values:
		# 1:width (1st term), 2:depth (2nd term), or 3:height (3rd term).
			BGL.glColor3f(*CL_TEXT_MD)
			BGL.glRasterPos2i( 504,  windowtop + ro - 98 )
			Text("Sort metrics data and standard names by ", 'normal')
			sortname = " %t|  width %x1|  depth %x2|  height %x3"
			object_mn_sort_by =		Menu(sortname,			19,	736, windowtop + ro - 104,	78, 22,	object_mn_sort_by.val,			"")

Register(gui, event, button_event)


